@@ -2,25 +2,37 @@
 # (c) 2023 knuxify and Ear Tag contributors
 
 from .config import config, DLCoverSize
-from .utils.bgtask import EartagBackgroundTask, run_threadsafe
+from .utils.asynctask import EartagAsyncTask
 from .utils.validation import is_valid_music_file, VALID_AUDIO_MIMES
+from .utils.tagselector import EartagTagSelectorButton  # noqa: F401
 from .dialogs import (
+    EartagErrorType,
+    EartagErrorDialog,
     EartagCloseWarningDialog,
     EartagDiscardWarningDialog,
     EartagTagDeleteWarningDialog,
 )
-from .musicbrainz import MusicBrainzRelease
 from .fileview import EartagFileView  # noqa: F401
 from .filemanager import EartagFileManager
 from .filelist import EartagFileList, EartagFileListItem  # noqa: F401
-from .rename import EartagRenameDialog
-from .identify import EartagIdentifyDialog
-from .extract import EartagExtractTagsDialog
 from . import APP_GRESOURCE_PATH, DEVEL
 
+import asyncio
 from gi.repository import Adw, Gdk, GLib, Gtk, Gio, GObject
 import os
-import gettext
+from enum import Flag
+
+
+class EartagFileDialogType(Flag):
+    """Types of file opening dialogs."""
+
+    # Dialog type - file or folder
+    TYPE_FILE = 1
+    TYPE_FOLDER = 2
+
+    # Load mode - overwrite (clear all and load) or insert (append to current)
+    MODE_OVERWRITE = 4
+    MODE_INSERT = 8
 
 
 @Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/nofile.ui")
@@ -29,11 +41,18 @@ class EartagNoFile(Adw.Bin):
 
     nofile_status = Gtk.Template.Child()
     open_file = Gtk.Template.Child()
+    open_recursive = Gtk.Template.Child()
 
     def __init__(self):
         super().__init__()
         if DEVEL:
             self.nofile_status.set_icon_name("app.drey.EarTag.Devel")
+        config.bind(
+            "open-folders-recursively",
+            self.open_recursive,
+            "active",
+            flags=Gio.SettingsBindFlags.DEFAULT,
+        )
 
     def grab_button_focus(self, *args):
         self.open_file.grab_focus()
@@ -41,14 +60,16 @@ class EartagNoFile(Adw.Bin):
     @Gtk.Template.Callback()
     def on_add_file(self, *args):
         window = self.get_native()
-        window.open_mode = EartagFileManager.LOAD_OVERWRITE
-        window.show_file_chooser(folders=False)
+        window.show_file_chooser(
+            EartagFileDialogType.TYPE_FILE | EartagFileDialogType.MODE_OVERWRITE
+        )
 
     @Gtk.Template.Callback()
     def on_add_folder(self, *args):
         window = self.get_native()
-        window.open_mode = EartagFileManager.LOAD_OVERWRITE
-        window.show_file_chooser(folders=True)
+        window.show_file_chooser(
+            EartagFileDialogType.TYPE_FOLDER | EartagFileDialogType.MODE_OVERWRITE
+        )
 
 
 @Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/window.ui")
@@ -80,8 +101,6 @@ class EartagWindow(Adw.ApplicationWindow):
     force_close = False
     file_chooser_mode = None
 
-    open_mode = EartagFileManager.LOAD_OVERWRITE
-
     # Sidebar
     sidebar_list_stack = Gtk.Template.Child()
     sidebar_list_scroll = Gtk.Template.Child()
@@ -103,6 +122,9 @@ class EartagWindow(Adw.ApplicationWindow):
     def __init__(self, application, paths=None, devel=False):
         super().__init__(application=application, title=_("Ear Tag"))
         self._undo_all_data = {}
+        self.error_dialog = None
+        self._identify_imported = False
+        self._redo_all_called = False
 
         if devel:
             self.add_css_class("devel")
@@ -110,26 +132,91 @@ class EartagWindow(Adw.ApplicationWindow):
         self.set_icon_name(application.get_application_id())
 
         self.file_chooser = Gtk.FileDialog(modal=True)
-        self._cancellable = Gio.Cancellable.new()
 
         self.audio_file_filter = Gtk.FileFilter()
         for mime in VALID_AUDIO_MIMES:
             self.audio_file_filter.add_mime_type(mime)
         self.audio_file_filter.set_name(_("All supported audio files"))
 
+        self.file_view.connect("notify::bind-in-progress", self.update_busy)
+
+        # File manager setup
         self.file_manager = EartagFileManager(self)
         self.file_view.set_file_manager(self.file_manager)
         self.file_manager.connect("notify::is-modified", self.toggle_save_button)
         self.file_manager.connect("notify::has-error", self.toggle_save_button)
-        self.file_manager.connect(
-            "notify::is-selected-modified", self.toggle_undo_all_action
-        )
+        self.file_manager.connect("notify::is-selected-modified", self.toggle_undo_all_action)
         self.file_manager.files.connect("items-changed", self.toggle_fileview)
         self.file_manager.connect("refresh-needed", self.update_state)
         self.file_manager.connect("selection-changed", self.update_state)
+        self.file_manager.connect("notify::is-busy", self.update_busy)
+        self.file_manager.connect("refresh-needed", self.refresh_actionbar_button_state)
+        self.file_manager.files.connect("items-changed", self.refresh_actionbar_button_state)
+        self.file_manager.connect("has-unwritable", self.show_unwritable_toast)
+        self.file_manager.connect("selection-changed", self.refresh_actionbar_button_state)
+
+        self.file_manager.load_task.connect("progress-pulse", self.loading_progressbar_pulse)
+        self.file_manager.load_task.connect("notify::progress", self.update_loading_progressbar)
         self.file_manager.load_task.connect(
-            "notify::progress", self.update_loading_progress
+            "task-done", self.task_done_handler, EartagErrorType.ERROR_LOAD
         )
+        self.file_manager.connect("max-recursion-depth-reached", self.show_max_recursion_toast)
+
+        self.file_manager.save_task.connect("notify::progress", self.update_loading_progressbar)
+        self.file_manager.save_task.connect(
+            "task-done", self.task_done_handler, EartagErrorType.ERROR_SAVE
+        )
+
+        self.file_manager.rename_task.connect(
+            "task-done", self.task_done_handler, EartagErrorType.ERROR_RENAME
+        )
+
+        self.show_error_action = Gio.SimpleAction.new(
+            "show-error-dialog", GLib.VariantType.new("(is)")
+        )
+        self.show_error_action.connect("activate", self._show_error_action)
+        self.add_action(self.show_error_action)
+
+        self.connect("close-request", self.on_close_request)
+
+        self.drop_target = Gtk.DropTarget(
+            actions=Gdk.DragAction.COPY,
+            formats=Gdk.ContentFormats.new_for_gtype(Gdk.FileList),
+        )
+        self.drop_target.connect("accept", self.on_drag_accept)
+        self.drop_target.connect("enter", self.on_drag_hover)
+        self.drop_target.connect("leave", self.on_drag_unhover)
+        self.drop_target.connect("drop", self.on_drag_drop)
+        self.add_controller(self.drop_target)
+
+        self.toggle_fileview()
+
+        # Tasks for undo/redo all option
+        self._undo_all_count = 0
+        self.undo_all_task = EartagAsyncTask(self._undo_all)
+        self.undo_all_task.connect("task-done", self._undo_all_done)
+
+        self._redo_all_count = 0
+        self.redo_all_task = EartagAsyncTask(self._redo_all)
+        self.redo_all_task.connect("task-done", self._redo_all_done)
+
+        # Task for delete all tags option
+        self._delete_all_tags_count = 0
+        self.delete_all_tags_task = EartagAsyncTask(self._delete_all_tags)
+        self.delete_all_tags_task.connect("task-done", self._delete_all_tags_done)
+
+        self._undo_delete_all_count = 0
+        self.undo_delete_all_tags_task = EartagAsyncTask(self._undo_delete_all_tags)
+        self.undo_delete_all_tags_task.connect("task-done", self._undo_delete_all_tags_done)
+
+        # Sidebar setup
+        self.sidebar_file_list.set_file_manager(self.file_manager)
+        self.sidebar_list_stack.set_visible_child(self.sidebar_no_files)
+
+        self.search_bar.set_key_capture_widget(self)
+        self.search_bar.connect_entry(self.search_entry)
+        self.search_entry.connect("search-changed", self.search_changed)
+
         self.search_bar.bind_property(
             "search-mode-enabled",
             self.sidebar_search_button,
@@ -148,86 +235,123 @@ class EartagWindow(Adw.ApplicationWindow):
             "revealed",
             GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
         )
-        self.sidebar_file_list.connect(
-            "notify::selection-mode", self.update_selection_mode
-        )
+        self.sidebar_file_list.connect("notify::selection-mode", self.update_selection_mode)
 
-        self.connect("close-request", self.on_close_request)
-
-        self.drop_target = Gtk.DropTarget(
-            actions=Gdk.DragAction.COPY,
-            formats=Gdk.ContentFormats.new_for_gtype(Gdk.FileList),
-        )
-        self.drop_target.connect("accept", self.on_drag_accept)
-        self.drop_target.connect("enter", self.on_drag_hover)
-        self.drop_target.connect("leave", self.on_drag_unhover)
-        self.drop_target.connect("drop", self.on_drag_drop)
-        self.add_controller(self.drop_target)
-
-        self.toggle_fileview()
-
-        # Tasks for undo/redo all option
-        self._undo_all_count = 0
-        self.undo_all_task = EartagBackgroundTask(self._undo_all)
-        self.undo_all_task.connect("task-done", self._undo_all_done)
-
-        self._redo_all_count = 0
-        self.redo_all_task = EartagBackgroundTask(self._redo_all)
-        self.redo_all_task.connect("task-done", self._redo_all_done)
-
-        # Task for delete all tags option
-
-        self._delete_all_tags_count = 0
-        self.delete_all_tags_task = EartagBackgroundTask(self._delete_all_tags)
-        self.delete_all_tags_task.connect("task-done", self._delete_all_tags_done)
-
-        self._undo_delete_all_count = 0
-        self.undo_delete_all_tags_task = EartagBackgroundTask(
-            self._undo_delete_all_tags
-        )
-        self.undo_delete_all_tags_task.connect(
-            "task-done", self._undo_delete_all_tags_done
-        )
-
-        # Sidebar setup
-        self.sidebar_file_list.set_file_manager(self.file_manager)
-        self.sidebar_list_stack.set_visible_child(self.sidebar_no_files)
-
-        self.search_bar.set_key_capture_widget(self)
-        self.search_bar.connect_entry(self.search_entry)
-        self.search_entry.connect("search-changed", self.search_changed)
-
-        self.file_manager.connect("refresh-needed", self.refresh_actionbar_button_state)
-        self.file_manager.files.connect(
-            "items-changed", self.refresh_actionbar_button_state
-        )
-        self.file_manager.connect(
-            "selection-changed", self.refresh_actionbar_button_state
-        )
-        self.file_manager.load_task.connect(
-            "notify::progress", self.update_loading_progressbar
-        )
         self.refresh_actionbar_button_state()
 
         if paths:
-            self.file_manager.load_files(paths, mode=EartagFileManager.LOAD_OVERWRITE)
+            self.file_manager.load_files(paths, overwrite=True)
 
-    def update_loading_progress(self, task, *args):
-        loading_progress = task.progress
-        is_loading = loading_progress != 0
-        self.sidebar_headerbar.set_sensitive(not is_loading)
-        if self.file_manager.files.get_n_items() == 0 and is_loading:
-            self.container_stack.set_visible_child(self.split_view)
+    @GObject.Property(type=bool, default=False)
+    def is_busy(self):
+        return self.file_manager.is_busy or self.file_view.bind_in_progress
+
+    def update_busy(self, *args):
+        """
+        Mark the window as busy or not busy based on whether the file manager
+        is currently processing a long operation (opening/saving/renaming files).
+
+        Also handles showing error dialogs.
+        """
+        self.notify("is-busy")
+        busy = self.is_busy
+
+        self.sidebar_headerbar.set_sensitive(not busy)
+        self.loading_progressbar_revealer.set_reveal_child(busy)
+        self.sidebar_file_list.set_sensitive(not self.file_manager.is_busy)
+        self.file_view.content_clamp.set_sensitive(not busy)
+        self.props.application.select_all_action.set_enabled(not busy)
+
+        if self.file_manager.load_task.is_running:
+            self.file_view.set_visible_child(self.file_view.loading)
+        else:
+            self.file_view.set_visible_child(self.file_view.content_stack)
+
+        if self.file_manager.files.get_n_items() == 0:
+            if busy:
+                self.container_stack.set_visible_child(self.split_view)
+            else:
+                self.container_stack.set_visible_child(self.no_file)
+
+    def show_error_dialog(self, error_type: EartagErrorType, error_message: str):
+        """Show an error dialog with logs."""
+        error_dialog = EartagErrorDialog(error_type, error_message)
+        error_dialog.present(self)
+
+    def _show_error_action(self, action: Gio.Action, data: GLib.Variant):
+        """Wrapper around show_error_dialog for the win.show-error-dialog option."""
+        error_type_int = data.get_child_value(0).get_int32()
+        error_message = data.get_child_value(1).get_string()
+        return self.show_error_dialog(EartagErrorType(error_type_int), error_message)
+
+    def task_done_handler(self, task, error_type: EartagErrorType):
+        """
+        Handle a long file manager task (opening, saving or renaming) being done.
+
+        The type of task that finished can be derived from the error_type parameter.
+        """
+
+        # If the task encountered an error, show an error dialog
+        if task.errors:
+            error_message = "\n---\n".join(task.errors)
+            if self.file_manager.files.get_n_items() == 0:
+                self.show_error_dialog(error_type, error_message)
+            else:
+                param = GLib.Variant.new_tuple(
+                    GLib.Variant.new_int32(int(error_type)),
+                    GLib.Variant.new_string(error_message),
+                )
+                toast = Adw.Toast.new(_("Some files failed to load"))
+                toast.props.button_label = _("More Information")
+                toast.props.action_name = "win.show-error-dialog"
+                toast.props.action_target = param
+                self.toast_overlay.add_toast(toast)
+
+        # Handle file loading not detecting any supported files
+        if error_type == EartagErrorType.ERROR_LOAD:
+            if self.file_manager.load_task.n_items == 0:
+                toast = Adw.Toast.new(_("No supported files found in opened folder"))
+                self.toast_overlay.add_toast(toast)
+
+        # Handle saving being done
+        if error_type == EartagErrorType.ERROR_SAVE:
+            saved_file_count = self.file_manager.save_task.n_items - len(
+                self.file_manager.load_task.errors
+            )
+            save_message = ngettext(
+                "Saved changes to 1 file",
+                "Saved changes to {n} files",
+                saved_file_count,
+            ).format(n=saved_file_count)
+
+            self.toast_overlay.add_toast(Adw.Toast.new(save_message))
+
+    def show_unwritable_toast(self, *args):
+        """Show a toast saying that some files are read-only."""
+        unwritable_msg = _("Some of the opened files are read-only; changes cannot be saved")  # noqa: E501
+        self.toast_overlay.add_toast(Adw.Toast.new(unwritable_msg))
+
+    def show_max_recursion_toast(self, *args):
+        """Show a toast saying that the maximum recursion depth has been reached."""
+        max_depth_msg = _("Maximum folder depth reached, some files may not be loaded")
+        self.toast_overlay.add_toast(Adw.Toast.new(max_depth_msg))
 
     def update_state(self, *args):
         app = self.get_application()
 
         # Set up the active view (hide fileview if there are no selected files)
         selected_files_count = self.file_manager.get_n_selected()
+
+        if selected_files_count == self.file_manager.get_n_files():
+            self.select_all_button.set_icon_name("edit-select-none-symbolic")
+            self.select_all_button.set_tooltip_text(_("Unselect all files"))
+        else:
+            self.select_all_button.set_icon_name("edit-select-all-symbolic")
+            self.select_all_button.set_tooltip_text(_("Select all files"))
+
         if selected_files_count <= 0:
             try:
                 for action in (
-                    app.sort_action,
                     app.rename_action,
                     app.extract_action,
                     app.identify_action,
@@ -240,11 +364,8 @@ class EartagWindow(Adw.ApplicationWindow):
             self.set_title(_("Ear Tag"))
             self.window_title.set_subtitle("")
             if self.file_manager.files:
-                self.file_view.content_stack.set_visible_child(
-                    self.file_view.select_file
-                )
+                self.file_view.content_stack.set_visible_child(self.file_view.select_file)
 
-            self.run_sort()
             return False
         else:
             files = self.file_manager.selected_files_list
@@ -274,7 +395,6 @@ class EartagWindow(Adw.ApplicationWindow):
         try:
             app.undo_all_action.set_enabled(is_modified)
             for action in (
-                app.sort_action,
                 app.rename_action,
                 app.extract_action,
                 app.identify_action,
@@ -286,17 +406,13 @@ class EartagWindow(Adw.ApplicationWindow):
 
         self.toggle_save_button()
 
-        self.file_view.update_binds()
-
     def toggle_fileview(self, *args):
         """
         Shows/hides the fileview/"no files" message depending on opened files.
         """
         if self.file_manager.files.get_n_items() > 0:
             self.container_stack.set_visible_child(self.split_view)
-            self.sidebar_headerbar.set_sensitive(
-                self.file_manager.load_task.progress in (0, 1)
-            )
+            self.sidebar_headerbar.set_sensitive(not self.is_busy)
             self.sidebar_list_stack.set_visible_child(self.sidebar_list_scroll)
         else:
             self.container_stack.set_visible_child(self.no_file)
@@ -336,10 +452,7 @@ class EartagWindow(Adw.ApplicationWindow):
 
     def on_drag_drop(self, drop_target, value, *args):
         files = value.get_files()
-
-        self.open_mode = EartagFileManager.LOAD_INSERT
-        self.load_files_from_paths([f.get_path() for f in files])
-        self.open_mode = EartagFileManager.LOAD_OVERWRITE
+        self.file_manager.load_files([f.get_path() for f in files], overwrite=False)
         self.on_drag_unhover()
 
     @Gtk.Template.Callback()
@@ -347,91 +460,67 @@ class EartagWindow(Adw.ApplicationWindow):
         """Sets the drop highlight to be invisible."""
         revealer.set_visible(revealer.props.child_revealed)
 
-    def show_file_chooser(self, folders=False):
-        """Shows the file chooser."""
-        if self.file_chooser_mode is not None:
-            return
+    async def show_file_chooser_async(self, chooser: EartagFileDialogType):
+        """Show the file chooser dialog and get files from it."""
+
+        # Set up title and check for chooser type validity
+        if chooser & EartagFileDialogType.TYPE_FILE:
+            if chooser & EartagFileDialogType.MODE_OVERWRITE:
+                #: TRANSLATORS: Title of the popup for opening a file
+                self.file_chooser.props.title = _("Open File")
+            elif chooser & EartagFileDialogType.MODE_INSERT:
+                self.file_chooser.props.title = _("Add File")
+            else:
+                raise ValueError
+
+        elif chooser & EartagFileDialogType.TYPE_FOLDER:
+            if chooser & EartagFileDialogType.MODE_OVERWRITE:
+                self.file_chooser.props.title = _("Open Folder")
+            else:
+                raise ValueError
 
-        if folders:
-            title = _("Open Folder")
-            self.file_chooser_mode = "folders"
         else:
-            title = _("Open File")
-            self.file_chooser_mode = "files"
+            raise ValueError
 
-        self.file_chooser.set_title(title)
-
-        if not folders:
+        # Set up filters for file selectors
+        if chooser & EartagFileDialogType.TYPE_FILE:
             _filters = Gio.ListStore.new(Gtk.FileFilter)
             _filters.append(self.audio_file_filter)
             self.file_chooser.set_filters(_filters)
-
-        if folders:
-            self.file_chooser.select_multiple_folders(
-                self, self._cancellable, self.open_file_from_dialog
-            )
         else:
-            self.file_chooser.open_multiple(
-                self, self._cancellable, self.open_file_from_dialog
-            )
+            self.file_chooser.set_filters(None)
 
-    def open_files(self, paths):
-        """
-        Loads the files with the given paths. Note that this does not perform
-        any validation; caller functions are meant to check for this manually.
-        """
-        if (
-            self.open_mode != EartagFileManager.LOAD_INSERT
-            and self.file_manager._is_modified
-        ):
-            self.discard_warning = EartagDiscardWarningDialog(self.file_manager, paths)
+        # If we're doing an overwrite load, warn about discarding changes
+        if chooser & EartagFileDialogType.MODE_OVERWRITE and self.file_manager._is_modified:
+            self.discard_warning = EartagDiscardWarningDialog()
             self.discard_warning.present(self)
-            return False
+            response = await self.discard_warning.wait_for_response()
 
-        self.file_manager.load_files(paths, mode=self.open_mode)
+            if response == "cancel":
+                del self.discard_warning
+                return
+            elif response == "save":
+                self.file_manager.save()
 
-    def open_file_from_dialog(self, dialog, result):
-        """
-        Callback for a FileChooser that takes the response and opens the file
-        selected in the dialog.
-        """
+            del self.discard_warning
+
+        # Display the dialog
         try:
-            if self.file_chooser_mode == "folders":
-                response = dialog.select_multiple_folders_finish(result)
-            elif self.file_chooser_mode == "files":
-                response = dialog.open_multiple_finish(result)
+            if chooser & EartagFileDialogType.TYPE_FOLDER:
+                gfiles = await self.file_chooser.select_multiple_folders(self, None)
             else:
-                self.file_chooser_mode = None
-                return False
+                gfiles = await self.file_chooser.open_multiple(self, None)
         except GLib.GError:
-            self.file_chooser_mode = None
-            return False
-
-        self.file_chooser_mode = None
-
-        if not response:
             return
 
-        return self.load_files_from_paths([f.get_path() for f in response])
+        await self.file_manager.load_files_async(
+            [f.get_path() for f in gfiles],
+            overwrite=bool(chooser & EartagFileDialogType.MODE_OVERWRITE),
+        )
 
-    def load_files_from_paths(self, paths):
-        """Loads files from a list of paths: handles both files and folders."""
-        paths_valid = []
-        for path in paths:
-            if os.path.isdir(path):
-                for file in os.listdir(path):
-                    fpath = os.path.join(path, file)
-                    if os.path.isfile(fpath) and is_valid_music_file(fpath):
-                        paths_valid.append(fpath)
-            else:
-                paths_valid.append(path)
-
-        if not paths_valid:
-            toast = Adw.Toast.new(_("No supported files found in opened folder"))
-            self.toast_overlay.add_toast(toast)
-            return
-
-        return self.open_files(paths_valid)
+    def show_file_chooser(self, chooser: EartagFileDialogType):
+        """Non-async wrapper around show_file_chooser_async."""
+        asyncio.create_task(self.show_file_chooser_async(chooser))
 
     def toggle_save_button(self, *args):
         if self.file_manager.has_error:
@@ -454,19 +543,13 @@ class EartagWindow(Adw.ApplicationWindow):
     def hide_sidebar(self, *args):
         self.split_view.set_show_sidebar(False)
 
-    def run_sort(self, *args):
-        if self.file_manager.files.get_n_items():
-            self.file_manager.sorter.changed(Gtk.SorterChange.DIFFERENT)
-
     @Gtk.Template.Callback()
     def on_save(self, *args):
-        if not self.file_manager.save():
-            return False
+        self.file_manager.save()
 
     @Gtk.Template.Callback()
     def insert_file(self, *args):
-        self.open_mode = EartagFileManager.LOAD_INSERT
-        self.show_file_chooser()
+        self.show_file_chooser(EartagFileDialogType.TYPE_FILE | EartagFileDialogType.MODE_INSERT)
 
     def on_close_request(self, *args):
         if (
@@ -474,13 +557,14 @@ class EartagWindow(Adw.ApplicationWindow):
             and list(self.file_manager.files)
             and self.file_manager._is_modified
         ):
-            self.close_request_dialog = EartagCloseWarningDialog(
-                self, self.file_manager
-            )
+            self.close_request_dialog = EartagCloseWarningDialog(self, self.file_manager)
             self.close_request_dialog.present(self)
             return True
 
-        MusicBrainzRelease.clear_tempfiles()
+        if self._identify_imported:
+            from .musicbrainz import EartagCAACover
+
+            EartagCAACover.clear_tempfiles()
 
         for file in self.file_manager.files:
             file.on_remove()
@@ -491,10 +575,15 @@ class EartagWindow(Adw.ApplicationWindow):
         self.file_view.on_close()
 
     def show_rename_dialog(self, *args):
+        from .rename import EartagRenameDialog
+
         self.rename_dialog = EartagRenameDialog(self)
         self.rename_dialog.present(self)
 
     def show_identify_dialog(self, *args):
+        from .identify import EartagIdentifyDialog
+
+        self._identify_imported = True
         self.identify_dialog = EartagIdentifyDialog(self)
         self.identify_dialog.present(self)
 
@@ -503,6 +592,8 @@ class EartagWindow(Adw.ApplicationWindow):
         self.preferences_dialog.present(self)
 
     def show_extract_dialog(self, *args):
+        from .extract import EartagExtractTagsDialog
+
         self.extract_dialog = EartagExtractTagsDialog(self)
         self.extract_dialog.present(self)
 
@@ -513,14 +604,12 @@ class EartagWindow(Adw.ApplicationWindow):
     # Sidebar
 
     def update_loading_progressbar(self, task, *args):
-        """
-        Updates the loading progressbar's position.
-        """
-        loading_progress = task.progress
-        self.loading_progressbar_revealer.set_reveal_child(not loading_progress == 0)
-        self.set_sensitive(loading_progress == 0)
-        self.sidebar_file_list.set_visible(loading_progress == 0)
-        self.loading_progressbar.set_fraction(loading_progress)
+        """Updates the loading progressbar's position."""
+        self.loading_progressbar.set_fraction(task.progress)
+
+    def loading_progressbar_pulse(self, task, *args):
+        """Updates the loading progressbar's pulse."""
+        self.loading_progressbar.pulse()
 
     def search_changed(self, search_entry, *args):
         """Emitted when the search has changed."""
@@ -532,11 +621,7 @@ class EartagWindow(Adw.ApplicationWindow):
         else:
             self.toggle_fileview()
 
-        if (
-            not self.selection_mode
-            and not self.file_manager.get_n_selected()
-            and n_results
-        ):
+        if not self.selection_mode and not self.file_manager.get_n_selected() and n_results:
             self.file_manager.select_file(self.file_manager.file_filter_model[0])
 
         # Scroll back to top of list
@@ -608,16 +693,10 @@ class EartagWindow(Adw.ApplicationWindow):
 
     def scroll_to_top(self):
         """Scrolls to the top of the file list."""
-        GLib.idle_add(self._scroll_to_top)
-
-    def _scroll_to_top(self):
         self.sidebar_list_scroll.get_vadjustment().set_value(0)
 
     def scroll_to_index(self, index):
         """Scrolls to file at the specified index."""
-        GLib.idle_add(self._scroll_to_index, index)
-
-    def _scroll_to_index(self, index):
         item_height = self.sidebar_file_list.get_first_child().get_height()
         vadjust = self.sidebar_list_scroll.get_vadjustment()
         new_value = item_height * (index + 1)
@@ -630,10 +709,9 @@ class EartagWindow(Adw.ApplicationWindow):
 
     def undo_all(self, *args):
         self.set_sensitive(False)
-        self.undo_all_task.reset()
         self.undo_all_task.run()
 
-    def _undo_all(self, *args):
+    async def _undo_all(self, *args):
         """Undo all changes and add the option to re-do them."""
         self._undo_all_data = {}
         self._undo_all_count = 0
@@ -646,16 +724,10 @@ class EartagWindow(Adw.ApplicationWindow):
             for tag in file.modified_tags:
                 self._undo_all_data[file.id][tag] = file.get_property(tag)
 
-            run_threadsafe(file.undo_all)
-
-        if self._undo_all_count == 0:
-            self.undo_all_task.emit_task_done()
-            return
-
-        self.undo_all_task.emit_task_done()
+            file.undo_all()
 
     def _undo_all_done(self, *args):
-        self.file_view.more_tags_group.slow_refresh_entries()
+        self._redo_all_called = False
         self.set_sensitive(True)
 
         toast = Adw.Toast.new(
@@ -669,15 +741,15 @@ class EartagWindow(Adw.ApplicationWindow):
         )
         toast.props.button_label = _("Redo")
         toast.connect("button-clicked", self.redo_all)
-        # toast.connect('dismissed', self.clear_redo_data)
+        toast.connect("dismissed", self.clear_redo_data)
         self.toast_overlay.add_toast(toast)
 
     def redo_all(self, *args):
+        self._redo_all_called = True
         self.set_sensitive(False)
-        self.redo_all_task.reset()
         self.redo_all_task.run()
 
-    def _redo_all(self, *args):
+    async def _redo_all(self, *args):
         """Reverses undo_all."""
         self._redo_all_count = 0
         for file in self.file_manager.files:
@@ -685,14 +757,11 @@ class EartagWindow(Adw.ApplicationWindow):
                 self._redo_all_count += 1
                 for tag, value in self._undo_all_data[file.id].items():
                     if tag in file.int_properties + file.float_properties and not value:
-                        run_threadsafe(file.set_property, tag, 0)
+                        file.set_property(tag, 0)
                     else:
-                        run_threadsafe(file.set_property, tag, value)
-
-        self.redo_all_task.emit_task_done()
+                        file.set_property(tag, value)
 
     def _redo_all_done(self, *args):
-        self.file_view.more_tags_group.slow_refresh_entries()
         self.set_sensitive(True)
         toast = Adw.Toast.new(
             # TRANSLATORS: {n} is a placeholder for the amount of files.
@@ -704,41 +773,39 @@ class EartagWindow(Adw.ApplicationWindow):
             ).format(n=self._redo_all_count)
         )
         self.toast_overlay.add_toast(toast)
+        self._redo_all_called = False
+        self.clear_redo_data()
 
     def clear_redo_data(self, *args):
         """Clears the redo data for the "undo all" option."""
-        self._undo_all_data = {}
+        if not self._redo_all_called:
+            del self._undo_all_data
+            self._undo_all_data = {}
 
     def toggle_undo_all_action(self, *args):
         app = self.get_application()
         try:
-            app.undo_all_action.set_enabled(
-                self.file_manager.props.is_selected_modified
-            )
+            app.undo_all_action.set_enabled(self.file_manager.props.is_selected_modified)
         except AttributeError:
             pass
 
     # "Delete all tags" option
     def do_delete_all_tags(self, *args):
         self.set_sensitive(False)
-        self.delete_all_tags_task.reset()
         self.delete_all_tags_task.run()
 
-    def _delete_all_tags(self, *args):
+    async def _delete_all_tags(self, *args):
         """Undo all changes and add the option to re-do them."""
         self._delete_all_tags_undo_data = {}
         files = self.file_manager.selected_files_list.copy()
         self._delete_all_tags_count = len(files)
         for file in files:
             self._delete_all_tags_undo_data[file.id] = {}
-            for prop in file.modified_tags:
+            for prop in file.present_tags:
                 self._delete_all_tags_undo_data[file.id][prop] = file.get_property(prop)
-            run_threadsafe(file.delete_all_raw)
-
-        self.delete_all_tags_task.emit_task_done()
+            file.delete_all_raw()
 
     def _delete_all_tags_done(self, *args):
-        self.file_view.more_tags_group.slow_refresh_entries()
         self.set_sensitive(True)
 
         toast = Adw.Toast.new(
@@ -758,23 +825,19 @@ class EartagWindow(Adw.ApplicationWindow):
 
     def undo_delete_all_tags(self, *args):
         self.set_sensitive(False)
-        self.undo_delete_all_tags_task.reset()
         self.undo_delete_all_tags_task.run()
 
-    def _undo_delete_all_tags(self, *args):
+    async def _undo_delete_all_tags(self, *args):
         """Reverses undo_all."""
         self._undo_delete_all_count = 0
         for file in self.file_manager.files:
             if file.id in self._delete_all_tags_undo_data:
                 self._undo_delete_all_count += 1
-                file.reload(thread_safe=True)
+                await file.reload()
                 for prop, value in self._delete_all_tags_undo_data[file.id].items():
-                    run_threadsafe(file.set_property, prop, value)
-
-        self.undo_delete_all_tags_task.emit_task_done()
+                    file.set_property(prop, value)
 
     def _undo_delete_all_tags_done(self, *args):
-        self.file_view.more_tags_group.slow_refresh_entries()
         self.set_sensitive(True)
 
         toast = Adw.Toast.new(
@@ -796,6 +859,7 @@ class EartagWindow(Adw.ApplicationWindow):
 class EartagPreferencesDialog(Adw.PreferencesDialog):
     __gtype_name__ = "EartagPreferencesDialog"
 
+    open_folders_recursively_toggle = Gtk.Template.Child()
     mb_confidence_spinbutton = Gtk.Template.Child()
     aid_confidence_spinbutton = Gtk.Template.Child()
     cover_size_comborow = Gtk.Template.Child()
@@ -803,6 +867,13 @@ class EartagPreferencesDialog(Adw.PreferencesDialog):
     def __init__(self):
         super().__init__()
 
+        config.bind(
+            "open-folders-recursively",
+            self.open_folders_recursively_toggle,
+            "active",
+            flags=Gio.SettingsBindFlags.DEFAULT,
+        )
+
         config.bind(
             "musicbrainz-confidence-treshold",
             self.mb_confidence_spinbutton,
